/*
	Blizzard - Heroes of the Storm - M3 importer
	
	Written by Taylor Mouse
*/

/* GLOBALS */
global mdlName = ""

global allBones = #()
global allSeqs = #()
global allIndices = #()
global allVertices = #()
global allNormals = #()
global allBoneWeights = #()
global allBoneIndices = #()
global allUVWs = #()
global allAttachments = #()
global allSubMeshes = #()

/*
	The most important part is the tag !
*/
struct SC2Tag 	  ( Name, OffSet, Length = 0, Flags )
struct SC2Bone 	  ( Id, Name,  ParentId, BoneTransform, BoneRef )
struct SC2Seqs	  ( Id, Name ) 
struct SC2Att	  ( Id, Name, ParentId, AttTransform, AttRef )
struct SC2Submesh ( Id, Name, firstVertex, nVerts, firstIndex, nIndices )

/*
	Read a specific number of bytes, without returning anything
*/
fn ReadBytes stream val =
(	
	for i=1 to val do ReadByte stream
)

/* 
	Read four bytes in reverse and return the string value
*/
fn ReadFourCC stream =
(
	ret = ""
	
	for i=1 to 4 do
	(
		ret = (bit.IntAsChar(ReadByte stream)) + ret
		
	)
	
	return ret
)

/* 
	Read a string defined with a fixed length
*/
fn ReadFixedString stream val =
(
	local str = ""
	for i=1 to val do
	(
		s= bit.IntAsChar(ReadByte stream)
		
		if(s!="\0") then str+=s
	)
	return str
)


/**************************************************************************************************************************/
/*  READ THE M3 */
/**************************************************************************************************************************/
fn Read =
(
	file = @"F:\StarCraft II Exported\Heroes Of The Storm\Extracted\Models\IllidanDemon.m3"
		
	stream = fOpen file "rb"

	id = ReadFourCC stream
	
	if( id != "MD33" and id != "MD34" ) then  return
	

	ofTag = 0
	ofTag = ReadLong stream #unsigned 
	nTag = ReadLong stream #unsigned
	entries = ReadLong stream #unsigned
	refId = ReadLong stream #unsigned
	
	tags = #()
	
	fSeek stream ofTag #seek_set
	
	for i=1 to nTag do
	(
		local tag = SC2Tag()
		tag.Name = ReadFourCC stream
		tag.Offset = ReadLong stream
		tag.Length = ReadLong stream
		tag.Flags = ReadLong stream
		
		append tags tag
	)
	
	for i=1 to nTag do
	(
		local tag = tags[i]
		if(tag.Name == "MODL") then
		(
			fSeek stream tags[i+1].OffSet #seek_set
			
			mdlName = GetFilenameFile ( ReadFixedString stream  tags[i+1].Length )
		) --// END MODL
		
		if(tag.Name == "BONE" ) then
		(
			for b=1 to tag.Length do
			(
				local aBone = SC2Bone()
				
				if( tags[i+b].Name == "CHAR") then
				(
					fSeek stream tags[i+b].OffSet #seek_set
					aBone.Name = ReadFixedString stream  tags[i+b].Length
					aBone.Id = b
				)
				append allBones aBone
			)
		) --// END BONE
		
		if(tag.Name =="SEQS" ) then
		(
			
			for s=1 to tag.Length do
			(
				local seq = SC2Seqs()			
				fSeek stream tags[i + s ].OffSet #seek_set
				seq.Id = s
				seq.Name = ReadFixedString stream  tags[i + s].Length
				
				append allSeqs seq
			)
		) -- // END SEQS
		
		if(tag.Name =="DIV_" ) then --// Indices
		(
			fSeek stream tag.OffSet #seek_set
			
			nIndices = ReadShort stream #unsigned
			size  = ReadLong  stream #unsigned
			
			for s=1 to tag.Length do
			(
				fSeek stream tags[i+s].OffSet #seek_set
				
				for t=1 to tags[i+s].Length do
				(
					-- f = [(( ReadShort stream #unsigned ) +1 ), (( ReadShort stream #unsigned ) +1 ), (( ReadShort stream #unsigned ) +1 ) ]
					append allIndices (( ReadShort stream #unsigned ) +1 )
				)

			)
		) -- // END INDICES
		
		if(tag.Name =="U8__" ) then --// Vertices
		(
			fSeek stream 128 #seek_set
		
			flag = 0
			chunkSize = 0
			flag = ReadLong stream #unsigned
			
			if(flag == 25297021) then chunkSize = 32
			if(flag == 25297533) then chunkSize = 36
			if(flag == 25296993) then chunkSize = 32
			if(flag == 25559649) then chunkSize = 40
			if(flag == 25297505) then chunkSize = 36
			if(flag == 25559677) then chunkSize = 40
			if(flag == 25559137) then chunkSize = 36
			if(flag == 25559165) then chunkSize = 36
			if(flag == 26083937) then chunkSize = 44
			if(flag == 26083453) then chunkSize = 40
			if(flag == 26083425) then chunkSize = 40
			if(flag == 27132001) then chunkSize = 44
			if(flag == 27132513) then chunkSize = 48
			if(flag == 1099038817) then chunkSize = 36
			if(flag == 1099300961) then chunkSize = 40
			if(flag == 1099038845) then chunkSize = 36
			if(flag == 1099300989) then chunkSize = 40
			if(flag == 27132029) then chunkSize = 44
			
			nVerts = tag.Length / chunkSize
			
			fSeek stream tag.OffSet #seek_set
			
			for v=1 to nVerts do
			(
				x1 = ReadFloat stream * 50
				y1 = ReadFloat stream * 50
				z1 = ReadFloat stream * 50
				
				append allVertices [ x1, y1, z1 ] 
				
				x2 = (Readbyte stream #unsigned) /255.0
				y2 = (Readbyte stream #unsigned) /255.0
				z2 = (Readbyte stream #unsigned) /255.0
				w2 = (Readbyte stream #unsigned) /255.0
				
				append allBoneWeights [ x2, y2, z2, w2 ] -- weight refers to the bone in the allBoneIndices
				
				x3 = (ReadByte stream #unsigned)
				y3 = (ReadByte stream #unsigned)
				z3 = (ReadByte stream #unsigned)
				w3 = (ReadByte stream #unsigned)
							
				append allBoneIndices 	[ x3, y3, z3, w3 ] -- max 4 bones per index
			
				x4 = (((ReadByte stream #unsigned ) /255.0) * 2 ) - 1
				y4 = (((ReadByte stream #unsigned ) /255.0) * 2 ) - 1
				z4 = (((ReadByte stream #unsigned ) /255.0) * 2 ) - 1
				w4 = (((ReadByte stream #unsigned ) /255.0) * 2 ) - 1

				append allNormals 		[ x4, y4, z4]
				
				uvCorrection = 2048.0
				u5 = (ReadShort stream #signed ) / uvCorrection
				v5 = 1.0 - ((ReadShort stream #signed ) / uvCorrection )
				w5 = 0.0 
				append allUVWs [ u5,v5,w5]
				
				ReadBytes stream (chunkSize - 28)
			)
		) -- // END VERTICES
		
		if(tag.Name == "ATT_" ) then -- // Attachment points
		(
			for b=1 to tag.Length do
			(
				local att = SC2Att()
				
				if( tags[i+b].Name == "CHAR") then
				(
					fSeek stream tags[i+b].OffSet #seek_set
					att.Id = b
					att.Name = ReadFixedString stream  tags[i+b].Length
				)
				append allAttachments att
			)
		) --// END ATTACHMENT POINTS
		
		
		if(tag.Name == "REGN" ) then -- // Submeshes
		(
			fSeek stream tag.OffSet #seek_set
			
			for s=1 to tag.Length do
			(
				local submesh = SC2Submesh()
				
				ReadBytes stream 4
				
				if( id != "MD33") then
				(
					ReadBytes stream 4
					submesh.firstVertex = (ReadLong stream #unsigned) as integer
					submesh.nVerts = (ReadLong stream #unsigned) as integer
				)
				else
				(
					submesh.firstVertex = (ReadShort stream #unsigned) as integer
					submesh.nVerts = (ReadShort stream #unsigned) as integer
				)
				submesh.Id = s
				submesh.Name = "Mesh " + (s as string)
				submesh.firstIndex = (ReadLong stream #unsigned) as integer
				submesh.nIndices = (ReadLong stream #unsigned) as integer
				
				ReadBytes stream 12
				
				append allSubMeshes submesh
			)
		) --// END SUBMESHES
		
		
		
	) -- // END TAGS

	fClose stream
)

/**************************************************************************************************************************/
/*  BUILD THE MESH */
/**************************************************************************************************************************/

fn Build =
(
	nMeshes = allSubMeshes.count
	nFirstFace = 0
	
	for m=1 to nMeshes do
	(
		submesh = allSubMeshes[m]
		print submesh
		
		local v = #()
		local n = #()
		local uv = #()

		for x=1 to submesh.nVerts do
		(
			append v allVertices[submesh.firstVertex + x]
			append n allNormals[submesh.firstVertex + x]
			append uv allUVWs[submesh.firstVertex + x]
		)

		local f = #()
		
		for x=1 to submesh.nIndices by 3 do
		(
			local face = [0,0,0]
			
			for j=1 to 3 do
			(
				local k = submesh.firstIndex + x + j - 1
				face[j] = allIndices[k]
			)
			append f face
		)
					
		nFirstFace = lastx
		
		local theMesh = mesh vertices:v vnorms:n name:submesh.Name tverts:uv faces:f
		
		
		theMesh.WireColor = Color (random 0 128) (random 0 255) (random 0 128)
		
		buildTVFaces theMesh false
		for i = 1 to f.count do
		( setTVFace theMesh i f[i] )
		
		update theMesh
	)
	
)

/**************************************************************************************************************************/
/*  CALL THE FUNCTIONS */
/**************************************************************************************************************************/
ClearListener()
Read()
Build()
